﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;

namespace Football
{
    class Utils
    {
        public enum BallState
        {
            weHaveBall,
            enemyHasBall,
            noOneHasBall
        }

        public enum TargetGroup
        {
            allPlayers,
            myPlayers,
            opositePlayers
        }

		protected Simulator s;
        protected TeamAI t;

        public Utils(Simulator sim, TeamAI team)
        {
            this.s = sim;
            this.t = team;
            initialize();
        }

        protected void initialize()
        {
            playground = s.GetPlayground();
            enemyGoal = s.GetOppositeGoal(t);
            MyPlayersIDs = s.getPlayersIDs(t);
            OpositePlayersIDs = s.getOpositePlayersIDs(t);
            AllPlayersIDs = new List<int>(MyPlayersIDs.Count + OpositePlayersIDs.Count);
            AllPlayersIDs.AddRange(MyPlayersIDs);
            AllPlayersIDs.AddRange(OpositePlayersIDs);

            EnemyGoalCentralPoint = new PointF(enemyGoal.X + enemyGoal.Width / 2, enemyGoal.Y + enemyGoal.Height / 2);
            Rectangle friendlyGoal = s.GetGoal(t);
            OurGoalCentralPoint = new PointF(friendlyGoal.X + friendlyGoal.Width / 2, friendlyGoal.Y + friendlyGoal.Height / 2);
            Locations = new Dictionary<int, PointF>(AllPlayersIDs.Count);
            Velocities = new Dictionary<int, PointF>(AllPlayersIDs.Count);
            MyPenaltyPoint = (s.GetOppositeGoal(this.t).X > 150) ? s.GetPlayground().firstPenaltyPoint : s.GetPlayground().secondPenaltyPoint;
            EnemyPenaltyPoint = (s.GetOppositeGoal(this.t).X > 150) ? s.GetPlayground().secondPenaltyPoint : s.GetPlayground().firstPenaltyPoint;
        }

        public void Update()
        {
            BallLocation.X = s.getBallLocation().X;
            BallLocation.Y = s.getBallLocation().Y;
            BallVelocity.X = s.getBallVelocity().X;
            BallVelocity.Y = s.getBallVelocity().Y;
            foreach (var item in AllPlayersIDs)
            {
                Locations[item] = s.getLocation(item);
                Velocities[item] = s.getVelocity(item);
            }
            determineBallState();
        }

        protected Playground playground;
        protected BallState ballState = BallState.noOneHasBall, previousBallState;
        public PointF EnemyGoalCentralPoint, OurGoalCentralPoint;
        protected Rectangle enemyGoal;
        
        public List<int> MyPlayersIDs, OpositePlayersIDs, AllPlayersIDs;
        /// <summary>
        /// Contains current locations of all players indexed by their ID
        /// </summary>
        public Dictionary<int, PointF> Locations;
        /// <summary>
        /// Contains current locations of all players indexed by their ID
        /// </summary>
        public Dictionary<int, PointF> Velocities;
        protected int ballHolder;
        public PointF BallLocation, BallVelocity;
        public PointF MyPenaltyPoint;
        public PointF EnemyPenaltyPoint;

        public float GetMaxPlayerVelocity()
        {
            return playground.maxPlayerSpeed;
        }

        public float GetMaxBallVelocity()
        {
            return playground.maxBallSpeed;
        }

        public double GetDistanceSquared(PointF from, PointF to)
        {
            return (to.X - from.X) * (to.X - from.X) + (to.Y - from.Y) * (to.Y - from.Y);
        }

		public double GetDistance(PointF from, PointF to)
		{
			return Math.Sqrt(GetDistanceSquared(from, to));
		}

		protected void determineBallState()
        {
            if (ballState == BallState.enemyHasBall || ballState == BallState.weHaveBall)
            {
                previousBallState = ballState;
            }
            ballState = BallState.noOneHasBall;
            foreach (var item in MyPlayersIDs)
            {
                if (s.HasBall(item))
                {
                    ballState = BallState.weHaveBall;
                    ballHolder = item;
                }
            }
            if (ballState != BallState.weHaveBall)
            {
                foreach (var item in OpositePlayersIDs)
                {
                    if (s.HasBall(item))
                    {
                        ballState = BallState.enemyHasBall;
                        ballHolder = item;
                    }
                }
            }
        }

        protected static PointF scale(PointF velocity, float velocityLimit)
        {
            float length = Math.Abs(velocity.X * velocity.X) + Math.Abs(velocity.Y * velocity.Y);
            if (length <= velocityLimit * velocityLimit)
                return velocity;
            else
            {
                float scaleMultiplicator = (float)(velocityLimit / Math.Sqrt(length));
                PointF result = new PointF(velocity.X * scaleMultiplicator, velocity.Y * scaleMultiplicator);
                return result;
            }
        }

        private static float evalNormalForm(float lineA, float lineB, float lineC, PointF point)
        {
            return lineA * point.X + lineB * point.Y + lineC;
        }

        private static float evalFromPoints(PointF lineStart, PointF normalVector, PointF testPoint)
        {
            float LineC = -normalVector.X * lineStart.X - normalVector.Y * lineStart.Y;
            return evalNormalForm(normalVector.X, normalVector.Y, LineC, testPoint);
        }

        private static bool isSameHalfplane(PointF lineStart, PointF normalVector, PointF testPoint1, PointF testPoint2)
        {
            float result1 = evalFromPoints(lineStart, normalVector, testPoint1),
                result2 = evalFromPoints(lineStart, normalVector, testPoint2);
            return (result1 <= 0 && result2 <= 0) || (result1 >= 0 && result2 >= 0);
        }

        private static bool isIn(PointF A, PointF B, PointF C, PointF D, PointF testPoint, PointF direction, PointF normal)
        {
            return isSameHalfplane(A, normal, testPoint, B) &&
                isSameHalfplane(B, normal, testPoint, A) &&
                isSameHalfplane(A, direction, testPoint, C) &&
                isSameHalfplane(C, direction, testPoint, A);
        }

        /// <summary>
        /// Tells you which side is in control of the ball
        /// </summary>
        /// <returns></returns>
        public BallState GetBallState()
        {
            determineBallState();
            return ballState;
        }

        /// <summary>
        /// Returns ID of the player currently controling the ball.
        /// </summary>
        /// <returns></returns>
        public int GetBallHoldersID()
        {
            determineBallState();
            return ballHolder;
        }

        /// <summary>
        /// Computes a vector of velocity required to move from the point <paramref name="from"/> to the point <paramref name="to"/>
        /// </summary>
        /// <param name="from"></param>
        /// <param name="to"></param>
        /// <returns></returns>
        public PointF ComputeVelocity(PointF from, PointF to)
        {
            return new PointF(10 * (to.X - from.X), 10 * (to.Y - from.Y)); ;
        }

        /// <summary>
        /// Returns ID of a player that is closest to the given location. Only some groups of players are consideres based on the parameter "target".
        /// </summary>
        /// <param name="spot"></param>
        /// <param name="target"></param>
        /// <returns></returns>
        public int GetNearestPlayer(PointF spot, TargetGroup target)
        {
            return GetNearestPlayer(spot, target, -1);
        }

        /// <summary>
        /// Returns ID of a player that is closest to the given location. Only some groups of players are consideres based on the parameter "target". The player whos ID equals to "exceptThisPlayer" is not considered.
        /// </summary>
        /// <param name="spot"></param>
        /// <param name="target"></param>
        /// <param name="exceptThisPlayer"></param>
        /// <returns></returns>
        public int GetNearestPlayer(PointF spot, TargetGroup target, int exceptThisPlayer)
        {
            int nearestPlayersID = -1;
            double nearestDistance = double.MaxValue;
            List<int> players = null;
            switch (target)
            {
                case TargetGroup.allPlayers:
                    players = AllPlayersIDs;
                    break;
                case TargetGroup.myPlayers:
                    players = MyPlayersIDs;
                    break;
                case TargetGroup.opositePlayers:
                    players = OpositePlayersIDs;
                    break;
            }
            for (int i = 0; i < players.Count; i++)
            {
                if (players[i] == exceptThisPlayer)
                    continue;
                double playersDistance = GetDistanceSquared(Locations[players[i]], spot);
                if (playersDistance < nearestDistance)
                {
                    nearestDistance = playersDistance;
                    nearestPlayersID = players[i];
                }
            }
            return nearestPlayersID;
        }

        /// <summary>
        /// Returns true if and only if there is a path from a place "from" to a place "to" of a width of "width" that doesn't contain any enemy players.
        /// </summary>
        /// <param name="from"></param>
        /// <param name="to"></param>
        /// <param name="width"></param>
        /// <returns></returns>
        protected bool isPathFree(PointF from, PointF to, float width)
        {
            PointF directionVector = new PointF(to.X - from.X, to.Y - from.Y),
                normalVector = new PointF(directionVector.Y, -directionVector.X);
            normalVector = scale(normalVector, width);
            PointF A = new PointF(from.X + normalVector.X, from.Y + normalVector.Y),
                B = new PointF(from.X - normalVector.X, from.Y - normalVector.Y),
                C = new PointF(to.X + normalVector.X, to.Y + normalVector.Y),
                D = new PointF(to.X - normalVector.X, to.Y - normalVector.Y);

            foreach (var item in OpositePlayersIDs)
            {
                PointF testPoint = Locations[item];
                if (isIn(A, B, C, D, testPoint, directionVector, normalVector))
                    return false;
            }
            return true;
        }


    }
}
